// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

import { argon2 } from 'argon2-browser';
import createHash from 'create-hash';
import createHmac from 'create-hmac';
import utils from './utils.js';

function getCurrentUrl(){
    if (typeof document === 'undefined' || !document)
        return null;

    var scripts = document.getElementsByTagName("script");
    var script = scripts[scripts.length - 1].src;

    return script.slice(0, script.lastIndexOf("/"));
}

let asmUrl = getCurrentUrl();

export default class Hash {
    static passwords(pass, salt, servers) {
        return Hash.argon2Buffer(pass, salt)
            .then(hash => servers.map(server => ({ server, hash: Hash.hmacBuffer(pass + server, hash).slice(-1) })))
            .then(hashes => hashes.map(item => ({server: item.server, pass: Buffer.from(item.hash).toString("base64").slice(0, 2)})));
    }

    static argon2(pass, salt) {
        return Hash.argon2Buffer(pass, salt).then(hash => hash.toString('base64'));
    }

    static argon2Buffer(pass, salt) {
        return argon2.hash({
            pass: pass,
            salt: salt,
            time: 4, // the number of iterations
            mem: 5120, // used memory, in KiB
            hashLen: 32, // desired hash length
            parallelism: 4, // desired parallelism (will be computed in parallel only for PNaCl)
            type: argon2.ArgonType.Argon2, // or argon2.ArgonType.Argon2i
            distPath: asmUrl // asm.js script location, without trailing slash
        }).then(data => Buffer.from(data.hash));
    }

    static sha(message) {
        return Hash.shaBuffer(message).toString('base64');
    }

    static shaBuffer(message) {
        return createHash('sha256').update(message).digest();
    }

    static hmac(data, secret) {
        return Hash.hmacBuffer(data, secret).toString('base64');
    }
    
    static hmacBuffer(data, secret) {
        secret = !utils.isBuffer(secret) ? Buffer.from(secret) : secret;
        return createHmac('sha256', secret).update(data).digest();
    }

    static get asmUrl() { return asmUrl; }
    static set asmUrl(value) { asmUrl = value; }
}