// Tide Protocol - Infrastructure for the Personal Data economy
// Copyright (C) 2019 Tide Foundation Ltd
// 
// This program is free software and is subject to the terms of 
// the Tide Community Open Source License as published by the 
// Tide Foundation Limited. You may modify it and redistribute 
// it in accordance with and subject to the terms of that License.
// This program is distributed WITHOUT WARRANTY of any kind, 
// including without any implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.
// See the Tide Community Open Source License for more details.
// You should have received a copy of the Tide Community Open 
// Source License along with this program.
// If not, see https://tide.org/licenses_tcosl-1-0-en

//best bignumber libraries jsbn big.js bn.js big-integer
import BigInt from 'big-integer';
import { randomBytes } from 'crypto';

export const constants = {
    zero: new BigInt(0),
    one: new BigInt(1),
    two: new BigInt(2)
};

export default class Utils {
    static encode(numbers, bits, encoding = undefined, fill = false) {
        const bytes = Utils.encodeByteArray(numbers, bits, fill);
        return Buffer.from(bytes).toString(encoding);
    }

    //TODO: Improve this to use Uint32Array and not waste to much memory
    static encodeByteArray(numbers, bits, fill = false) {
        const blockSize = Math.ceil(bits / 8);
        const arrayBuffer = [];

        for (let i = 0; i < numbers.length; i++) {
            const bytes = numbers[i].toArray(256).value;

            if (fill) { //fill the empty bits
                const diff = blockSize - bytes.length;
                for (let j = 0; j < diff; j++)
                    bytes.unshift(0);
            }

            arrayBuffer.push(bytes);
        }

        return Array.prototype.concat.apply([], arrayBuffer);
    }

    //TODO: Doesn't work with chinese characters, like text = "I'm 大戴禮記, and my dad's ابو عبد الله🤣";
    static decode(data, bits, encoding = undefined) {
        if (!Utils.isBuffer(data))
            data = Buffer.from(data, encoding);
    
        const blockSize = bits / 8;
        let numbers = [];
        let j = -1;
    
        for (let i = 0; i < data.length; i++) {
            if (i % blockSize === 0) {
                j++
                numbers.push(constants.zero)
            }
    
            numbers[j] = numbers[j].shiftLeft(8).or(new BigInt(data[i].toString()));
        }
        return numbers;
    }
    
    static getPrimitiveRoot(p) {
        if (p.equals(constants.two))
            return constants.one;
    
        const min = new BigInt(3); // Avoid g=2 because of Bleichenbacher's attack
        const max = p.subtract(constants.one);
        const p1 = constants.two; // The prime divisors of p-1 are 2 and (p-1)/2
        const p2 = max.divide(p1); // Because: p = 2q + 1 where q is a prime
    
        while (true) {
            const g = this.random(min, max);
            if (!g.modPow(p1, p).equals(constants.one) &&
                !g.modPow(p2, p).equals(constants.one) &&
                !p.subtract(constants.one).mod(g).equals(constants.zero) && // g|p-1
                !p.subtract(constants.one).mod(g.modInv(p)).equals(constants.zero)) // g^(-1)|p-1 (evades Khadir's attack)
                return g
        }
    }
    
    static random(min, max) {
        const rang = max.subtract(min);
        const rdmLen = Math.ceil(rang.bitLength() / 8)
        const diff = (rdmLen * 8) - rang.bitLength();
    
        let rdm;
        do {
            rdm = new BigInt(randomBytes(rdmLen).toString('hex'), 16)
            rdm = rdm.shiftRight(diff).add(min);
        } while (rdm.compareTo(max) > 0);
    
        return rdm;
    }
    
    //min == 2 ^ (bits-2) == 1 << bits-2 that is, 1 and bits-2 zeros to the right
    //max == ((2 ^ bits) -2)/2 that is, all the bits in 1 with bits-1 length
    //because q = (p -1) / 2 and p <= (2 ^ bits) -1 that is, all the bits in 1
    static randomPrime(bits, confidence) {
        const maxLen = bits - 2;
        const rdmLen = Math.ceil(maxLen / 8);
        const diff = (rdmLen * 8) - maxLen;
        const min = constants.one.shiftLeft(bits - 2);
    
        while (true) {
            const rdm = new BigInt(randomBytes(rdmLen).toString('hex'), 16);
            //trunk to bits, add minimum range and the number has to be odd
            const q = rdm.shiftRight(diff).add(min).or(constants.one);
            if (!q.isProbablePrime(confidence))
                continue;
    
            // p = (q * 2) + 1
            const p = q.shiftLeft(1).add(constants.one);
            if (!p.isProbablePrime(confidence))
                continue;
    
            return [p, q];
        }
    }

    static isBuffer(data) {
        return Buffer.isBuffer(data) || ArrayBuffer.isView(data)
    }
}
